import 'dart:math' as math;

import 'nullable.dart';
import 'list.dart';
import 'types.dart';
import 'union.dart';

/// A class to validate number types
class AcanthisNumber extends AcanthisType<num> {
  AcanthisNumber();

  /// Add a check to the number to check if it is less than or equal to [value]
  AcanthisNumber lte(num value) {
    addCheck(AcanthisCheck<num>(
        onCheck: (toTest) => toTest <= value,
        error: 'Value must be less than or equal to $value',
        name: 'lte'));
    return this;
  }

  /// Add a check to the number to check if it is greater than or equal to [value]
  AcanthisNumber gte(num value) {
    addCheck(AcanthisCheck<num>(
        onCheck: (toTest) => toTest >= value,
        error: 'Value must be greater than or equal to $value',
        name: 'gte'));
    return this;
  }

  /// Add a check to the number to check if it is greater than [value]
  AcanthisNumber gt(num value) {
    addCheck(AcanthisCheck<num>(
        onCheck: (toTest) => toTest > value,
        error: 'Value must be greater than $value',
        name: 'gt'));
    return this;
  }

  /// Add a check to the number to check if it is less than [value]
  AcanthisNumber lt(num value) {
    addCheck(AcanthisCheck<num>(
        onCheck: (toTest) => toTest < value,
        error: 'Value must be less than $value',
        name: 'lt'));
    return this;
  }

  /// Add a check to the number to check if it is positive
  AcanthisNumber positive() {
    addCheck(AcanthisCheck<num>(
        onCheck: (toTest) => toTest > 0,
        error: 'Value must be positive',
        name: 'positive'));
    return this;
  }

  /// Add a check to the number to check if it is negative
  AcanthisNumber negative() {
    addCheck(AcanthisCheck<num>(
        onCheck: (toTest) => toTest < 0,
        error: 'Value must be negative',
        name: 'negative'));
    return this;
  }

  /// Add a check to the number to check if it is an integer
  AcanthisNumber integer() {
    addCheck(AcanthisCheck<num>(
        onCheck: (toTest) => toTest % 1 == 0,
        error: 'Value must be an integer',
        name: 'integer'));
    return this;
  }

  /// Add a check to the number to check if it is a multiple of [value]
  AcanthisNumber multipleOf(int value) {
    addCheck(AcanthisCheck<num>(
        onCheck: (toTest) => toTest % value == 0,
        error: 'Value must be a multiple of $value',
        name: 'multipleOf'));
    return this;
  }

  /// Add a check to the number to check if it is finite
  AcanthisNumber finite() {
    addCheck(AcanthisCheck<num>(
        onCheck: (toTest) => toTest.isFinite,
        error: 'Value is not finite',
        name: 'finite'));
    return this;
  }

  /// Add a check to the number to check if it is infinite
  AcanthisNumber infinite() {
    addCheck(AcanthisCheck<num>(
        onCheck: (toTest) => toTest.isInfinite,
        error: 'Value is not infinite',
        name: 'infinite'));
    return this;
  }

  /// Add a check to the number to check if it is "not a number"
  AcanthisNumber nan() {
    addCheck(AcanthisCheck<num>(
        onCheck: (toTest) => toTest.isNaN,
        error: 'Value is not NaN',
        name: 'nan'));
    return this;
  }

  /// Add a check to the number to check if it is not "not a number"
  AcanthisNumber notNaN() {
    addCheck(AcanthisCheck<num>(
        onCheck: (toTest) => !toTest.isNaN,
        error: 'Value is NaN',
        name: 'notNaN'));
    return this;
  }

  /// Create a list of numbers
  AcanthisList<num> list() {
    return AcanthisList<num>(this);
  }

  /// Add a custom check to the number
  AcanthisNumber customCheck(
      {required bool Function(num value) onCheck,
      required String error,
      required String name}) {
    addCheck(AcanthisCheck<num>(onCheck: onCheck, error: error, name: name));
    return this;
  }

  /// Transform the number to a power of [value]
  AcanthisNumber pow(int value) {
    addTransformation(AcanthisTransformation<num>(
        transformation: (toTransform) => math.pow(toTransform, value)));
    return this;
  }

  /// Add a transformation to the number to add [value]
  AcanthisNumber transform(num Function(num value) transformation) {
    addTransformation(
        AcanthisTransformation<num>(transformation: transformation));
    return this;
  }

  /// Make the value nullable
  AcanthisNullable<num> nullable({num? defaultValue}) {
    return AcanthisNullable(this, defaultValue: defaultValue);
  }

  /// Create a union from the number
  AcanthisUnion or(List<AcanthisType> elements) {
    return AcanthisUnion([this, ...elements]);
  }
}

/// Create a number type
AcanthisNumber number() => AcanthisNumber();
